from .client import FeathrClient
from .spark_provider.feathr_configurations import SparkExecutionConfiguration
from .definition.feature_derivations import *
from .definition.anchor import *
from .definition.feature import *
from .definition.feathrconfig import *
from .definition.transformation import *
from .definition.dtype import *
from .definition.source import *
from .definition.typed_key import *
from .definition.materialization_settings import *
from .definition.monitoring_settings import *
from .definition.sink import *
from .definition.query_feature_list import *
from .definition.lookup_feature import *
from .definition.aggregation import *
from .definition.settings import *
from .utils.job_utils import *
from .utils.feature_printer import *
from .version import __version__

# skipped class as they are internal methods:
# RepoDefinitions, HoconConvertible,
# expose the modules so docs can build
# referencee： https://stackoverflow.com/questions/15115514/how-do-i-document-classes-without-the-module-name/31594545#31594545

# __all__ = []
# for v in dir():
#     if not v.startswith('__') and v != 'mypackage':
#         __all__.append(v)


__all__ = [
    'FeatureJoinJobParams',
    'FeatureGenerationJobParams',
    'FeathrClient',
    'DerivedFeature',
    'FeatureAnchor',
    'Feature',
    'ValueType',
    'WindowAggTransformation',
    'TypedKey',
    'DUMMYKEY',
    'BackfillTime',
    'MaterializationSettings',
    'MonitoringSettings',
    'RedisSink',
    'HdfsSink',
    'MonitoringSqlSink',
    'AerospikeSink',
    'FeatureQuery',
    'LookupFeature',
    'Aggregation',
    'get_result_df',
    'AvroJsonSchema',
    'Source',
    'InputContext',
    'HdfsSource',
    'KafkaConfig',
    'KafKaSource',
    'ValueType',
    'BooleanFeatureType',
    'Int32FeatureType',
    'Int64FeatureType',
    'FloatFeatureType',
    'DoubleFeatureType',
    'StringFeatureType',
    'BytesFeatureType',
    'FloatVectorFeatureType',
    'Int32VectorFeatureType',
    'Int64VectorFeatureType',
    'DoubleVectorFeatureType',
    'FeatureNameValidationError',
    'ObservationSettings',
    'FeaturePrinter',
    'SparkExecutionConfiguration',
    __version__,
 ]



import inspect
from re import sub
from typing import List
from urllib.parse import urlparse
from feathr.constants import INPUT_CONTEXT
from feathr.definition.anchor import FeatureAnchor
from feathr.definition.dtype import FeatureType, str_to_value_type, value_type_to_str
from feathr.definition.feature import Feature
from feathr.definition.feature_derivations import DerivedFeature
from feathr.definition.source import HdfsSource, JdbcSource, Source
from pyapacheatlas.core import AtlasProcess,AtlasEntity

from feathr.definition.transformation import ExpressionTransformation, Transformation, WindowAggTransformation
from feathr.definition.typed_key import TypedKey
def to_camel(s):
    if not s:
        return s
    if isinstance(s, str):
        if "_" in s:
            s = sub(r"(_)+", " ", s).title().replace(" ", "")
            return ''.join([s[0].lower(), s[1:]])
        return s
    elif isinstance(s, list):
        return [to_camel(i) for i in s]
    elif isinstance(s, dict):
        return dict([(to_camel(k), s[k]) for k in s])
                                  
def source_to_def(v: Source) -> dict:
    # Note that after this method, attributes are Camel cased (eventTimestampColumn). 
    # If the old logic works with snake case (event_timestamp_column), make sure you handle them manually. 
    ret = {}
    if v.name == INPUT_CONTEXT:
        return {
            "name": INPUT_CONTEXT,
            "type": INPUT_CONTEXT,
            "path": INPUT_CONTEXT,
        }
    elif isinstance(v, HdfsSource):
        ret = {
            "name": v.name,
            "type": "hdfs",
            "path": v.path,
        }
    elif isinstance(v, JdbcSource):
        ret = {
            "name": v.name,
            "type": "jdbc",
            "url": v.url,
        }
        if hasattr(v, "dbtable") and v.dbtable:
            ret["dbtable"] = v.dbtable
        if hasattr(v, "query") and v.query:
            ret["query"] = v.query
        if hasattr(v, "auth") and v.auth:
            ret["auth"] = v.auth
    else:
        raise ValueError(f"Unsupported source type {v.__class__}")
    if hasattr(v, "preprocessing") and v.preprocessing:
        ret["preprocessing"] = inspect.getsource(v.preprocessing)
    if v.event_timestamp_column:
        ret["eventTimestampColumn"] = v.event_timestamp_column
        ret["event_timestamp_column"] = v.event_timestamp_column
    if v.timestamp_format:
        ret["timestampFormat"] = v.timestamp_format
        ret["timestamp_format"] = v.timestamp_format
    if v.registry_tags:
        ret["tags"] = v.registry_tags
    return ret
    
def anchor_to_def(v: FeatureAnchor) -> dict:
    # Note that after this method, attributes are Camel cased (eventTimestampColumn). 
    # If the old logic works with snake case (event_timestamp_column), make sure you handle them manually. 
    source_id = v.source._registry_id
    ret = {
        "name": v.name,
        "sourceId": str(source_id),
    }
    if v.registry_tags:
        ret["tags"] = v.registry_tags
    return ret

def transformation_to_def(v: Transformation) -> dict:
    if isinstance(v, ExpressionTransformation):
        return {
            "transformExpr": v.expr
        }
    elif isinstance(v, WindowAggTransformation):
        ret = {
            "defExpr": v.def_expr,
        }
        if v.agg_func:
            ret["aggFunc"] = v.agg_func
        if v.window:
            ret["window"] = v.window
        if v.group_by:
            ret["groupBy"] = v.group_by
        if v.filter:
            ret["filter"] = v.filter
        if v.limit:
            ret["limit"] = v.limit
        return ret
    raise ValueError("Unsupported Transformation type")

def feature_type_to_def(v: FeatureType) -> dict:
    # Note that after this method, attributes are Camel cased (eventTimestampColumn). 
    # If the old logic works with snake case (event_timestamp_column), make sure you handle them manually. 
    return {
        "type": v.type,
        "tensorCategory": v.tensor_category,
        "dimensionType": [value_type_to_str(t) for t in v.dimension_type],
        "valType": value_type_to_str(v.val_type),
    }

def typed_key_to_def(v: TypedKey) -> dict:
    ret = {
        "keyColumn": v.key_column,
        "keyColumnType": value_type_to_str(v.key_column_type)
    }
    if v.full_name:
        ret["fullName"] = v.full_name
    if v.description:
        ret["description"] = v.description
    if v.key_column_alias:
        ret["keyColumnAlias"] = v.key_column_alias
    return ret

def feature_to_def(v: Feature) -> dict:
    ret = {
        "name": v.name,
        "featureType": feature_type_to_def(v.feature_type),
        "key": [typed_key_to_def(k) for k in v.key],
    }
    if v.transform:
        ret["transformation"] = transformation_to_def(
            v.transform)
    if v.registry_tags:
        ret["tags"] = v.registry_tags
    return ret

def derived_feature_to_def(v: DerivedFeature) -> dict:
    # Note that after this method, attributes are Camel cased (eventTimestampColumn). 
    # If the old logic works with snake case (event_timestamp_column), make sure you handle them manually. 
    ret = {
        "name": v.name,
        "featureType": feature_type_to_def(v.feature_type),
        "key": [typed_key_to_def(k) for k in v.key],
        "inputAnchorFeatures": [str(f._registry_id) for f in v.input_features if not isinstance(f, DerivedFeature)],
        "inputDerivedFeatures": [str(f._registry_id) for f in v.input_features if isinstance(f, DerivedFeature)],
    }
    if v.transform:
        ret["transformation"] = transformation_to_def(v.transform)
    return ret

def topological_sort(derived_feature_list: List[DerivedFeature]) -> List[DerivedFeature]:
    """
    In the current registry implementation, we need to make sure all upstream are registered before registering one derived feature
    Thus we need to sort derived features by the partial order of dependencies, upstream to downstream.
    """
    ret = []
    # We don't want to destroy the input list
    input = derived_feature_list.copy()
    
    # Each round add the most downstream features into `ret`, so `ret` is in reversed order
    while input:
        # Process all remaining features
        current = input.copy()
        
        # In Python you should not alter content while iterating
        current_copy = current.copy()
        
        # Go over all remaining features to see if some feature depends on others
        for f in current_copy:
            for i in f.input_features:
                if i in current:
                    # Someone depends on feature `i`, so `i` is **not** the most downstream
                    current.remove(i)
        
        # Now `current` contains only the most downstream features in this round
        ret.extend(current)
        
        # Remove one level of dependency from input
        for f in current:
            input.remove(f)
    
    # The ret was in a reversed order when it's generated
    ret.reverse()
    
    if len(set(ret)) != len (set(derived_feature_list)):
        raise ValueError("Cyclic dependency detected")
    return ret


from datetime import datetime
import json
import os
from pathlib import Path
from shlex import split
from subprocess import STDOUT, Popen
import time
from typing import Any, Dict, List, Optional

from loguru import logger
from pyspark import *

from feathr.constants import FEATHR_MAVEN_ARTIFACT
from feathr.spark_provider._abc import SparkJobLauncher


class _FeathrLocalSparkJobLauncher(SparkJobLauncher):
    """Class to interact with local Spark. This class is not intended to be used in Production environments.
    It is intended to be used for testing and development purposes. No authentication is required to use this class.

    Args:
        workspace_path (str): Path to the workspace
    """

    def __init__(
        self,
        workspace_path: str,
        master: str = None,
        debug_folder: str = "debug",
        clean_up: bool = True,
        retry: int = 3,
        retry_sec: int = 5,
    ):
        """Initialize the Local Spark job launcher"""
        self.workspace_path = (workspace_path,)
        self.debug_folder = debug_folder
        self.spark_job_num = 0
        self.clean_up = clean_up
        self.retry = retry
        self.retry_sec = retry_sec
        self.packages = self._get_default_package()
        self.master = master or "local[*]"

    def upload_or_get_cloud_path(self, local_path_or_http_path: str):
        """For Local Spark Case, no need to upload to cloud workspace."""
        return local_path_or_http_path

    def submit_feathr_job(
        self,
        job_name: str,
        main_jar_path: str,
        main_class_name: str,
        arguments: List[str] = None,
        python_files: List[str] = None,
        configuration: Dict[str, str] = {},
        properties: Dict[str, str] = {},
        **_,
    ) -> Any:
        """Submits the Feathr job to local spark, using subprocess args.
        Note that the Spark application will automatically run on YARN cluster mode. You cannot change it if
        you are running with Azure Synapse.

        Args:
            job_name: name of the job
            main_jar_path: main file paths, usually your main jar file
            main_class_name: name of your main class
            arguments: all the arguments you want to pass into the spark job
            python_files: required .zip, .egg, or .py files of spark job
            configuration: Additional configs for the spark job
            properties: System properties configuration
            **_: Not used arguments in local spark mode, such as reference_files_path and job_tags
        """
        logger.warning(
            f"Local Spark Mode only support basic params right now and should be used only for testing purpose."
        )
        self.cmd_file, self.log_path = self._get_debug_file_name(self.debug_folder, prefix=job_name)

        # Get conf and package arguments
        cfg = configuration.copy() if configuration else {}
        maven_dependency = f"{cfg.pop('spark.jars.packages', self.packages)},{FEATHR_MAVEN_ARTIFACT}"
        spark_args = self._init_args(job_name=job_name, confs=cfg)

        if not main_jar_path:
            # We don't have the main jar, use Maven
            if not python_files:
                # This is a JAR job
                # Azure Synapse/Livy doesn't allow JAR job starts from Maven directly, we must have a jar file uploaded.
                # so we have to use a dummy jar as the main file.
                logger.info(f"Main JAR file is not set, using default package '{FEATHR_MAVEN_ARTIFACT}' from Maven")
                # Use the no-op jar as the main file
                # This is a dummy jar which contains only one `org.example.Noop` class with one empty `main` function
                # which does nothing
                current_dir = Path(__file__).parent.resolve()
                main_jar_path = os.path.join(current_dir, "noop-1.0.jar")
                spark_args.extend(["--packages", maven_dependency, "--class", main_class_name, main_jar_path])
            else:
                spark_args.extend(["--packages", maven_dependency])
                # This is a PySpark job, no more things to
                if python_files.__len__() > 1:
                    spark_args.extend(["--py-files", ",".join(python_files[1:])])
                print(python_files)
                spark_args.append(python_files[0])
        else:
            spark_args.extend(["--class", main_class_name, main_jar_path])

        if arguments:
            spark_args.extend(arguments)

        if properties:
            spark_args.extend(["--system-properties", json.dumps(properties)])

        cmd = " ".join(spark_args)

        log_append = open(f"{self.log_path}_{self.spark_job_num}.txt", "a")
        proc = Popen(split(cmd), shell=False, stdout=log_append, stderr=STDOUT)
        logger.info(f"Detail job stdout and stderr are in {self.log_path}.")

        self.spark_job_num += 1

        with open(self.cmd_file, "a") as c:
            c.write(" ".join(proc.args))
            c.write("\n")

        self.latest_spark_proc = proc

        logger.info(f"Local Spark job submit with pid: {proc.pid}.")

        return proc

    def wait_for_completion(self, timeout_seconds: Optional[float] = 500) -> bool:
        """This function track local spark job commands and process status.
        Files will be write into `debug` folder under your workspace.
        """
        logger.info(f"{self.spark_job_num} local spark job(s) in this Launcher, only the latest will be monitored.")
        logger.info(f"Please check auto generated spark command in {self.cmd_file} and detail logs in {self.log_path}.")

        proc = self.latest_spark_proc
        start_time = time.time()
        retry = self.retry

        log_read = open(f"{self.log_path}_{self.spark_job_num-1}.txt", "r")
        while proc.poll() is None and (((timeout_seconds is None) or (time.time() - start_time < timeout_seconds))):
            time.sleep(1)
            try:
                if retry < 1:
                    logger.warning(
                        f"Spark job has hang for {self.retry * self.retry_sec} seconds. latest msg is {last_line}. \
                            Please check {log_read.name}"
                    )
                    if self.clean_up:
                        self._clean_up()
                        proc.wait()
                    break
                last_line = log_read.readlines()[-1]
                retry = self.retry
                if last_line == []:
                    print("_", end="")
                else:
                    print(">", end="")
                    if last_line.__contains__("Feathr Pyspark job completed"):
                        logger.info(f"Pyspark job Completed")
                        proc.terminate()
            except IndexError as e:
                print("x", end="")
                time.sleep(self.retry_sec)
                retry -= 1

        job_duration = time.time() - start_time
        log_read.close()

        if proc.returncode == None:
            logger.warning(
                f"Spark job with pid {self.latest_spark_proc.pid} not completed after {timeout_seconds} sec \
                    time out setting. Please check."
            )
            if self.clean_up:
                self._clean_up()
                proc.wait()
                return True
        elif proc.returncode == 1:
            logger.warning(f"Spark job with pid {self.latest_spark_proc.pid} is not successful. Please check.")
            return False
        else:
            logger.info(
                f"Spark job with pid {self.latest_spark_proc.pid} finished in: {int(job_duration)} seconds \
                    with returncode {proc.returncode}"
            )
            return True

    def _clean_up(self, proc: Popen = None):
        logger.warning(f"Terminate the spark job due to as clean_up is set to True.")
        if not proc:
            self.latest_spark_proc.terminate()
        else:
            proc.terminate()

    def get_status(self) -> str:
        """Get the status of the job, only a placeholder for local spark"""
        return self.latest_spark_proc.returncode

    def _init_args(self, job_name: str, confs: Dict[str, str]) -> List[str]:
        logger.info(f"Spark job: {job_name} is running on local spark with master: {self.master}.")
        args = [
            "spark-submit",
            "--master",
            self.master,
            "--name",
            job_name,
            "--conf",
            "spark.hadoop.fs.wasbs.impl=org.apache.hadoop.fs.azure.NativeAzureFileSystem",
            "--conf",
            "spark.hadoop.fs.wasbs=org.apache.hadoop.fs.azure.NativeAzureFileSystem",
        ]

        for k, v in confs.items():
            args.extend(["--conf", f"{k}={v}"])

        return args

    def _get_debug_file_name(self, debug_folder: str = "debug", prefix: str = None):
        """Auto generated command will be write into cmd file.
        Spark job output will be write into log path with job number as suffix.
        """
        prefix += datetime.now().strftime("%Y%m%d%H%M%S")
        debug_path = os.path.join(debug_folder, prefix)

        print(debug_path)
        if not os.path.exists(debug_path):
            os.makedirs(debug_path)

        cmd_file = os.path.join(debug_path, f"command.sh")
        log_path = os.path.join(debug_path, f"log")

        return cmd_file, log_path

    def _get_default_package(self):
        # default packages of Feathr Core, requires manual update when new dependency introduced or package updated.
        # TODO: automate this process, e.g. read from pom.xml
        # TODO: dynamical modularization: add package only when it's used in the job, e.g. data source dependencies.
        packages = []
        packages.append("org.apache.spark:spark-avro_2.12:3.3.0")
        packages.append("com.microsoft.sqlserver:mssql-jdbc:10.2.0.jre8")
        packages.append("com.microsoft.azure:spark-mssql-connector_2.12:1.2.0")
        packages.append("org.apache.logging.log4j:log4j-core:2.17.2,com.typesafe:config:1.3.4")
        packages.append("com.fasterxml.jackson.core:jackson-databind:2.12.6.1")
        packages.append("org.apache.hadoop:hadoop-mapreduce-client-core:2.7.7")
        packages.append("org.apache.hadoop:hadoop-common:2.7.7")
        packages.append("org.apache.hadoop:hadoop-azure:3.2.0")
        packages.append("org.apache.avro:avro:1.8.2,org.apache.xbean:xbean-asm6-shaded:4.10")
        packages.append("org.apache.spark:spark-sql-kafka-0-10_2.12:3.1.3")
        packages.append("com.microsoft.azure:azure-eventhubs-spark_2.12:2.3.21")
        packages.append("org.apache.kafka:kafka-clients:3.1.0")
        packages.append("com.google.guava:guava:31.1-jre")
        packages.append("it.unimi.dsi:fastutil:8.1.1")
        packages.append("org.mvel:mvel2:2.2.8.Final")
        packages.append("com.fasterxml.jackson.module:jackson-module-scala_2.12:2.13.3")
        packages.append("com.fasterxml.jackson.dataformat:jackson-dataformat-yaml:2.12.6")
        packages.append("com.fasterxml.jackson.dataformat:jackson-dataformat-csv:2.12.6")
        packages.append("com.jasonclawson:jackson-dataformat-hocon:1.1.0")
        packages.append("com.redislabs:spark-redis_2.12:3.1.0")
        packages.append("org.apache.xbean:xbean-asm6-shaded:4.10")
        packages.append("com.google.protobuf:protobuf-java:3.19.4")
        packages.append("net.snowflake:snowflake-jdbc:3.13.18")
        packages.append("net.snowflake:spark-snowflake_2.12:2.10.0-spark_3.2")
        packages.append("org.apache.commons:commons-lang3:3.12.0")
        packages.append("org.xerial:sqlite-jdbc:3.36.0.3")
        packages.append("com.github.changvvb:jackson-module-caseclass_2.12:1.1.1")
        packages.append("com.azure.cosmos.spark:azure-cosmos-spark_3-1_2-12:4.11.1")
        packages.append("org.eclipse.jetty:jetty-util:9.3.24.v20180605")
        packages.append("commons-io:commons-io:2.6")
        packages.append("org.apache.hadoop:hadoop-azure:2.7.4")
        packages.append("com.microsoft.azure:azure-storage:8.6.4")
        return ",".join(packages)



import string

import pytest

from feathr import FeatureBase


@pytest.mark.parametrize('bad_feature_name',
                         [None,
                          ''])
def test_feature_name_fails_on_empty_name(bad_feature_name: str):
    with pytest.raises(Exception, match="empty feature name"):
        FeatureBase.validate_feature_name(bad_feature_name)


def test_feature_name_fails_on_leading_number():
    with pytest.raises(Exception, match="cannot start with a number"):
        FeatureBase.validate_feature_name("4featurename")


def test_feature_name_fails_on_punctuation_chars():
    for char in set(string.punctuation) - set('_'):
        with pytest.raises(Exception,  match="only letters, numbers, and underscores are allowed"):
            FeatureBase.validate_feature_name(f"feature_{char}_name")


@pytest.mark.parametrize('feature_name',
                         ["feature_name",
                          "features4lyfe",
                          "f_4_feature_",
                          "_leading_underscores_are_ok",
                          "CapitalizedFeature"
                          ''])
def test_feature_name_validates_ok(feature_name: str):
    assert FeatureBase.validate_feature_name(feature_name)




import enum
from typing import Type, Union, List, Optional
from abc import ABC, abstractmethod
from jinja2 import Template
from feathr.definition.feathrconfig import HoconConvertible

class Transformation(HoconConvertible):
    """Base class for all transformations that produce feature values."""
    pass


class RowTransformation(Transformation):
    """Base class for all row-level transformations."""
    pass


class ExpressionTransformation(RowTransformation):
    """A row-level transformations that is defined with the Feathr expression language.

    Attributes:
        expr: expression that transforms the raw value into a new value, e.g. amount * 10.
    """
    def __init__(self, expr: str) -> None:
        super().__init__()
        self.expr = expr

    def to_feature_config(self, with_def_field_name: Optional[bool] = True) -> str:
        tm = Template("""
{% if with_def_field_name %}
def.sqlExpr: "{{expr}}"
{% else %}
"{{expr}}"
{% endif %}
        """)
        return tm.render(expr=self.expr, with_def_field_name=with_def_field_name)


class WindowAggTransformation(Transformation):
    """Aggregate the value of an expression over a fixed time window. E.g. sum(amount*10) over last 3 days.

    Attributes:
        agg_expr: expression that transforms the raw value into a new value, e.g. amount * 10
        agg_func: aggregation function. Available values: `SUM`, `COUNT`, `MAX`, `MIN`, `AVG`, `MAX_POOLING`, `MIN_POOLING`, `AVG_POOLING`, `LATEST`
        window: Time window length to apply the aggregation. support 4 type of units: d(day), h(hour), m(minute), s(second). The example value are "7d' or "5h" or "3m" or "1s"
        group_by: Feathr expressions applied after the `agg_expr` transformation as groupby field, before aggregation, same as 'group by' in SQL
        filter: Feathr expression applied to each row as a filter before aggregation. This should be a string and a valid Spark SQL Expression. For example: filter = 'age > 3'. This is similar to PySpark filter operation and more details can be learned here: https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.filter.html
    """
    def __init__(self, agg_expr: str, agg_func: str, window: str, group_by: Optional[str] = None, filter: Optional[str] = None, limit: Optional[int] = None) -> None:
        super().__init__()
        self.def_expr = agg_expr
        self.agg_func = agg_func
        self.window = window
        self.group_by = group_by
        self.filter = filter
        self.limit = limit

    def to_feature_config(self, with_def_field_name: Optional[bool] = True) -> str:
        tm = Template("""
def:"{{windowAgg.def_expr}}"
window: {{windowAgg.window}}
aggregation: {{windowAgg.agg_func}}
{% if windowAgg.group_by is not none %}
    groupBy: {{windowAgg.group_by}}
{% endif %}
{% if windowAgg.filter is not none %}
    filter: {{windowAgg.filter}}
{% endif %}
{% if windowAgg.limit is not none %}
    limit: {{windowAgg.limit}}
{% endif %}
        """)
        return tm.render(windowAgg = self)


class UdfTransform(Transformation):
    """User defined transformation. To be supported.

    Attributes:
        name: name of the user defined function
    """
    def __init__(self, name: str) -> None:
        """

        :param name:
        """
        super().__init__()
        self.name = name
    def to_feature_config(self) -> str:
        pass


import ast
import inspect
import os
import pickle
from pathlib import Path
from typing import List, Optional, Union

from jinja2 import Template

from feathr.definition.anchor import FeatureAnchor
from feathr.definition.source import HdfsSource


# Some metadata that are only needed by Feathr
FEATHR_PYSPARK_METADATA = 'generated_feathr_pyspark_metadata'
# UDFs that are provided by users and persisted by Feathr into this file.
# It will be uploaded into the Pyspark cluster as a dependency for execution
FEATHR_CLIENT_UDF_FILE_NAME = 'client_udf_repo.py'
# Pyspark driver code that is executed by the Pyspark driver
FEATHR_PYSPARK_DRIVER_FILE_NAME = 'feathr_pyspark_driver.py'
FEATHR_PYSPARK_DRIVER_TEMPLATE_FILE_NAME = 'feathr_pyspark_driver_template.py'
# Feathr provided imports for pyspark UDFs all go here
PROVIDED_IMPORTS = ['\nfrom pyspark.sql import SparkSession, DataFrame\n'] + \
                   ['from pyspark.sql.functions import *\n']


class _PreprocessingPyudfManager(object):
    """This class manages Pyspark UDF preprocessing related artifacts, like UDFs from users, the pyspark_client etc.
    """
    @staticmethod
    def build_anchor_preprocessing_metadata(anchor_list: List[FeatureAnchor], local_workspace_dir):
        """When the client build features, UDFs and features that need preprocessing will be stored as metadata. Those
        metadata will later be used when uploading the Pyspark jobs.
        """
        # feature names concatenated to UDF map
        # for example, {'f1,f2,f3': my_udf1, 'f4,f5':my_udf2}
        feature_names_to_func_mapping = {}
        # features that have preprocessing defined. This is used to figure out if we need to kick off Pyspark
        # preprocessing for requested features.
        features_with_preprocessing = []
        client_udf_repo_path = os.path.join(local_workspace_dir, FEATHR_CLIENT_UDF_FILE_NAME)
        metadata_path = os.path.join(local_workspace_dir, FEATHR_PYSPARK_METADATA)
        pyspark_driver_path = os.path.join(local_workspace_dir, FEATHR_PYSPARK_DRIVER_FILE_NAME)

        # delete the file if it already exists to avoid caching previous results
        for f in [client_udf_repo_path, metadata_path,  pyspark_driver_path]:
            if os.path.exists(f):
                os.remove(f)

        for anchor in anchor_list:
            # only support batch source preprocessing for now.
            if not hasattr(anchor.source, "preprocessing"):
                continue
            preprocessing_func = anchor.source.preprocessing
            if preprocessing_func:
                _PreprocessingPyudfManager.persist_pyspark_udf_to_file(preprocessing_func, local_workspace_dir)
                feature_names = [feature.name for feature in anchor.features]
                features_with_preprocessing = features_with_preprocessing + feature_names
                feature_names.sort()
                string_feature_list = ','.join(feature_names)
                if isinstance(anchor.source.preprocessing, str):
                    feature_names_to_func_mapping[string_feature_list] = _PreprocessingPyudfManager._parse_function_str_for_name(anchor.source.preprocessing)
                else:
                    # it's a callable function
                    feature_names_to_func_mapping[string_feature_list] = anchor.source.preprocessing.__name__

        if not features_with_preprocessing:
            return

        _PreprocessingPyudfManager.write_feature_names_to_udf_name_file(feature_names_to_func_mapping, local_workspace_dir)

        # Save necessary preprocessing-related metadata locally in your workspace
        # Typically it's used as a metadata for join/gen job to figure out if there is preprocessing UDF
        # exist for the features requested
        feathr_pyspark_metadata_abs_path = os.path.join(local_workspace_dir, FEATHR_PYSPARK_METADATA)
        with open(feathr_pyspark_metadata_abs_path, 'wb') as file:
            pickle.dump(features_with_preprocessing, file)


    @staticmethod
    def _parse_function_str_for_name(fn_str: str) -> str:
        """Use AST to parse the function string and get the name out.

        Args:
            fn_str: Function code in string.

        Returns:
            Name of the function.
        """
        if not fn_str:
            return None

        tree = ast.parse(fn_str)

        # tree.body contains a list of function definition objects parsed from the input string.
        # Currently, we only accept a single function.
        if len(tree.body) != 1 or not isinstance(tree.body[0], ast.FunctionDef):
            raise ValueError("provided code fragment is not a single function")

        # Get the function name from the function definition.
        return tree.body[0].name


    @staticmethod
    def persist_pyspark_udf_to_file(user_func, local_workspace_dir):
        """persist the pyspark UDF to a file in `local_workspace_dir` for later usage.
        The user_func could be either a string that represents a function body, or a callable object.
        The reason being - if we are defining a regular Python function, it will be a callable object;
        however if we retrieve features from registry, the current implementation is to use plain strings to store the function body. In that case, the user_fuc will be string.
        """
        if isinstance(user_func, str):
            udf_source_code = [user_func]
        else:
            udf_source_code = inspect.getsourcelines(user_func)[0]
        lines = []
        # Some basic imports will be provided
        lines = lines + PROVIDED_IMPORTS
        lines = lines + udf_source_code
        lines.append('\n')

        client_udf_repo_path = os.path.join(local_workspace_dir, FEATHR_CLIENT_UDF_FILE_NAME)

        # the directory may actually not exist yet, so create the directory first
        Path(local_workspace_dir).mkdir(parents=True, exist_ok=True)

        # Append to file, Create it if doesn't exist
        with open(client_udf_repo_path, "a+") as handle:
            print("".join(lines), file=handle)

    @staticmethod
    def write_feature_names_to_udf_name_file(feature_names_to_func_mapping, local_workspace_dir):
        """Persist feature names(sorted) of an anchor to the corresponding preprocessing function name to source path
        under the local_workspace_dir.
        """
        # indent in since python needs correct indentation
        # Don't change the indentation
        tm = Template("""
feature_names_funcs = {
{% for key, value in func_maps.items() %}
    "{{key}}" : {{value}},
{% endfor %}
}
        """)
        new_file = tm.render(func_maps=feature_names_to_func_mapping)

        full_file_name = os.path.join(local_workspace_dir, FEATHR_CLIENT_UDF_FILE_NAME)
        # Append to file, Create it if doesn't exist
        with open(full_file_name, "a+") as text_file:
            print(new_file, file=text_file)

    @staticmethod
    def prepare_pyspark_udf_files(feature_names: List[str], local_workspace_dir):
        """Prepare the Pyspark driver code that will be executed by the Pyspark cluster.
        The Pyspark driver code file has two parts:
            1. The driver code itself
            2. The UDFs.
            3. The features that need preprocessing in a map(from feature names to UDFs)
        """
        py_udf_files = []

        # Load pyspark_metadata which stores what features contains preprocessing UDFs
        feathr_pyspark_metadata_abs_path = os.path.join(local_workspace_dir, FEATHR_PYSPARK_METADATA)
        # if the preprocessing metadata file doesn't exist or is empty, then we just skip
        if not Path(feathr_pyspark_metadata_abs_path).is_file():
            return py_udf_files
        with open(feathr_pyspark_metadata_abs_path, 'rb') as pyspark_metadata_file:
            features_with_preprocessing = pickle.load(pyspark_metadata_file)
        # if there is not features that needs preprocessing, just return.
        if not features_with_preprocessing:
            return py_udf_files

        # Figure out if we need to preprocess via UDFs for requested features.
        # Only if the requested features contain preprocessing logic, we will load Pyspark. Otherwise just use Scala
        # spark.
        has_py_udf_preprocessing = False
        for feature_name in feature_names:
            if feature_name in features_with_preprocessing:
                has_py_udf_preprocessing = True

        if has_py_udf_preprocessing:
            pyspark_driver_path = os.path.join(local_workspace_dir, FEATHR_PYSPARK_DRIVER_FILE_NAME)
            pyspark_driver_template_abs_path = str(Path(Path(__file__).parent / FEATHR_PYSPARK_DRIVER_TEMPLATE_FILE_NAME).absolute())
            client_udf_repo_path = os.path.join(local_workspace_dir, FEATHR_CLIENT_UDF_FILE_NAME)
            # write pyspark_driver_template_abs_path and then client_udf_repo_path
            filenames = [pyspark_driver_template_abs_path, client_udf_repo_path]

            with open(pyspark_driver_path, 'w') as outfile:
                for fname in filenames:
                    with open(fname) as infile:
                        for line in infile:
                            outfile.write(line)
            lines = [
                '\n',
                'print("pyspark_client.py: Preprocessing via UDFs and submit Spark job.")\n',
                'submit_spark_job(feature_names_funcs)\n',
                'print("pyspark_client.py: Feathr Pyspark job completed.")\n',
                '\n',
            ]
            with open(pyspark_driver_path, "a") as handle:
                print("".join(lines), file=handle)

            py_udf_files = [pyspark_driver_path]
        return py_udf_files



OUTPUT_PATH_TAG = "output_path"
# spark config for output format setting
OUTPUT_FORMAT = "spark.feathr.outputFormat"
REDIS_PASSWORD = 'REDIS_PASSWORD'

# 1MB = 1024*1024
MB_BYTES = 1048576

INPUT_CONTEXT="PASSTHROUGH"
RELATION_CONTAINS = "CONTAINS"
RELATION_BELONGSTO = "BELONGSTO"
RELATION_CONSUMES = "CONSUMES"
RELATION_PRODUCES = "PRODUCES"
# For use in registry. 
# For type definition, think it's like a schema of a table. 
# This version field is mainly to smooth possible future upgrades, 
# for example, backward incompatible changes should be introduced in v2, to make sure that features registered with v1 schema can still be used
REGISTRY_TYPEDEF_VERSION="v1"

TYPEDEF_SOURCE=f'feathr_source_{REGISTRY_TYPEDEF_VERSION}'
# TODO: change the name from feathr_workspace_ to feathr_project_
TYPEDEF_FEATHR_PROJECT=f'feathr_workspace_{REGISTRY_TYPEDEF_VERSION}'
TYPEDEF_DERIVED_FEATURE=f'feathr_derived_feature_{REGISTRY_TYPEDEF_VERSION}'
TYPEDEF_ANCHOR=f'feathr_anchor_{REGISTRY_TYPEDEF_VERSION}'
TYPEDEF_ANCHOR_FEATURE=f'feathr_anchor_feature_{REGISTRY_TYPEDEF_VERSION}'

TYPEDEF_ARRAY_ANCHOR=f"array<feathr_anchor_{REGISTRY_TYPEDEF_VERSION}>"
TYPEDEF_ARRAY_DERIVED_FEATURE=f"array<feathr_derived_feature_{REGISTRY_TYPEDEF_VERSION}>"
TYPEDEF_ARRAY_ANCHOR_FEATURE=f"array<feathr_anchor_feature_{REGISTRY_TYPEDEF_VERSION}>"

# Decouple Feathr MAVEN Version from Feathr Python SDK Version
import os
from feathr.version import __version__
FEATHR_MAVEN_VERSION = os.environ.get("FEATHR_MAVEN_VERSION", __version__)
FEATHR_MAVEN_ARTIFACT=f"com.linkedin.feathr:feathr_2.12:{FEATHR_MAVEN_VERSION}"

JOIN_CLASS_NAME="com.linkedin.feathr.offline.job.FeatureJoinJob"
GEN_CLASS_NAME="com.linkedin.feathr.offline.job.FeatureGenJob"


import click
from py4j.java_gateway import JavaGateway
import pathlib
import os
from pathlib import Path
import distutils.dir_util
import subprocess
import urllib.request
from feathr.client import FeathrClient
from feathr.registry._feathr_registry_client import _FeatureRegistry

@click.group()
@click.pass_context
def cli(ctx: click.Context):
    """
    Feathr CLI tool. Visit https://github.com/feathr-ai/feathr for more details.
    """
    pass


def check_user_at_root():
    """
    Checks if the user is running the CLI from the root of the user workspace.

    We require the user to be running the CLI from the root of the user workspace so that path-related functionalities
    can work correctly.
    """
    # use this file as a anchor point to identify the root of the repo
    anchor_file = 'feathr_config.yaml'
    user_workspace_dir = Path(".")
    anchor_file_path = user_workspace_dir / anchor_file
    if not anchor_file_path.exists():
        raise click.UsageError('You are NOT at the root of your user workspace("/feathr_user_workspace"). Please '
                               'execute the command under your user workspace root.')


@cli.command()
@click.option('--name', default="feathr_user_workspace", help='Specify the workspace name.')
@click.option('--git/--no-git', default=False, help='When enabled, a git-based workspace will be created.')
def init(name, git):
    """
    Initializes a Feathr project to create and manage features. A team should share a same Feathr project usually via
    git. By default, git is NOT used to manage the workspace.
    """
    # Check if workspace already exists
    workspace_dir = Path(os.path.join(Path.cwd(), name))
    workspace_exist = os.path.isdir(workspace_dir)
    if workspace_exist:
        # workspace already exist. Just exit.
        raise click.UsageError(f'Feathr workspace ({name}) already exist. Please use a new folder name.')

    output_str = f'Creating workspace {name} with sample config files and mock data ...'
    click.echo(output_str)
    default_workspace = str(Path(Path(__file__).parent / 'data' / 'feathr_user_workspace').absolute())

    # current feathr_user_workspace directory w.r.t. where the init command is executed
    pathlib.Path(name).mkdir(parents=True, exist_ok=True)

    # create some default folders and example Feathr configs
    distutils.dir_util.copy_tree(default_workspace, str(workspace_dir))

    # Create a git repo for the workspace
    if git:
        os.chdir(workspace_dir)
        process = subprocess.Popen(['git', 'init'], stdout=subprocess.PIPE)
        output = process.communicate()[0]
        click.echo(output)
        click.echo(click.style('Git init completed for your workspace. Please read the '
                               'wiki to learn how to manage '
                               'your workspace with git.', fg='green'))
    click.echo(click.style('Feathr initialization completed.', fg='green'))


@cli.command()
@click.option('--save_to', default="./", help='Specify the path to save the output HOCON config(relative to current path).')
def hocon(save_to):
    """
    Scan all Python-based feature definitions recursively under current directory,
     convert them to HOCON config and save to a given path (relative to current directory).
    """
    scan_dir = Path.cwd()
    save_to = Path(os.path.join(scan_dir, save_to))
    _FeatureRegistry.save_to_feature_config(scan_dir, save_to)


@cli.command()
@click.argument('filepath', default='feature_join_conf/feature_join.conf', type=click.Path(exists=True))
def join(filepath):
    """
    Creates the offline training dataset with the requested features.
    """

    check_user_at_root()

    click.echo(click.style('Batch joining features with config: ' + filepath, fg='green'))
    with open(filepath) as f:
        lines = []
        for line in f:
            click.echo(line, nl=False)
            lines.append(line)
    click.echo()
    click.echo()

    client = FeathrClient()
    client._get_offline_features_with_config(filepath)
    click.echo(click.style('Feathr feature join job submitted. Visit '
                           'https://ms.web.azuresynapse.net/en-us/monitoring/sparkapplication for detailed job '
                           'result.', fg='green'))

@cli.command()
@click.argument('filepath', default='feature_gen_conf/feature_gen.conf', type=click.Path(exists=True))
def deploy(filepath):
    """
    Deploys the features to online store based on the feature generation config.
    """
    check_user_at_root()

    click.echo(click.style('Deploying feature generation config: ' + filepath, fg='green'))
    with open(filepath) as f:
        lines = []
        for line in f:
            click.echo(line, nl=False)
            lines.append(line)
    click.echo()
    click.echo()

    client = FeathrClient()
    client._materialize_features_with_config(filepath)
    click.echo()
    click.echo(click.style('Feathr feature deployment submitted. Visit '
                           'https://ms.web.azuresynapse.net/en-us/monitoring/sparkapplication for detailed job '
                           'result.', fg='green'))


@cli.command()
@click.option('--git/--no-git', default=False, help='If git-enabled, the new changes will be added and committed.')
@click.option('--msg', help='The feature name.')
def register(git, msg):
    """
    Register your feature metadata to your metadata registry.
    """

    check_user_at_root()

    # The register command is not integrated with Azure Atlas yet.
    click.echo(click.style('Registering your metadata to metadata service...', fg='green'))

    if git:
        click.echo(click.style('Git: adding all files.', fg='green'))
        click.echo(msg)
        process = subprocess.Popen(['git', 'add', '-A'], stdout=subprocess.PIPE)
        output = process.communicate()[0]
        click.echo(output)

        click.echo(click.style('Git: committing.', fg='green'))
        process2 = subprocess.Popen(['git', 'commit', '-m', msg], stdout=subprocess.PIPE)
        output2 = process2.communicate()[0]
        click.echo(output2)
    client = FeathrClient()
    client.register_features()
    click.echo(click.style('Feathr registration completed successfully!', fg='green'))


@cli.command()
def start():
    """
    Starts a local Feathr engine for local experimentation and testing.

    Feathr local test requires feathr local engine, a java jar, to stay running locally. You can download the jar
    yourself from feathr website or use this command to download. The jar should be placed under the root of the
    feathr_user_workspace. After the jar is downloaded, the command will run this jar. The jar needs to be running(
    don't close the terminal) while you want to use 'feathr test'.
    """
    def run_jar():
        cmd = ['java', '-jar', jar_name]
        with subprocess.Popen(cmd, stdout=subprocess.PIPE, bufsize=1, universal_newlines=True) as p:
            # Need to continuously pump the results from jar to terminal
            for line in p.stdout:
                print(line, end='')


    check_user_at_root()
    # The jar should be placed under the root of the user workspace
    jar_name = 'feathr_local_engine.jar'

    # Download the jar if it doesn't exist
    if not os.path.isfile(jar_name):
        url = 'https://azurefeathrstorage.blob.core.windows.net/public/' + jar_name
        file_name = url.split('/')[-1]
        u = urllib.request.urlopen(url)
        f = open(file_name, 'wb')
        meta = u.info()
        file_size = int(meta.get('Content-Length'))
        click.echo(click.style('There is no local feathr engine(jar) in the workspace. Will download the feathr jar.',
                               fg='green'))
        click.echo('Downloading feathr jar for local testing: %s Bytes: %s from %s' % (file_name, file_size, url))

        file_size_dl = 0
        block_sz = 8192
        with click.progressbar(length=file_size,
                               label='Download feathr local engine jar') as bar:
            while True:
                buffer = u.read(block_sz)
                if not buffer:
                    break

                file_size_dl += len(buffer)
                f.write(buffer)
                bar.update(block_sz)

        f.close()

    click.echo(click.style(f'Starting the local feathr engine: {jar_name}.'))
    click.echo(click.style(f'Please keep this open and start another terminal to run feathr test. This terminal shows '
                           f'the debug message.', fg='green'))

    run_jar()


@cli.command()
@click.option('--features', prompt='Your feature names, separated by comma', help='The feature name.')
def test(features):
    """
    Tests a single feature definition locally via local spark mode with mock data. Mock data has to be provided by the
    users. Please execute "feathr start" before "feathr test" to setup the local engine.
    """
    check_user_at_root()
    click.echo('\nProducing feature values for requested features ... ')

    gateway = JavaGateway()
    # User should run this command at user workspace dir root
    user_workspace_dir = os.path.abspath(".")

    # for py4j, it's always named as entry_point
    stack_entry_point_result = gateway.entry_point.getResult(user_workspace_dir, features)

    click.echo('\nFeature computation completed.')
    click.echo(stack_entry_point_result)



import os
import yaml
from loguru import logger
from feathr.secrets.akv_client import AzureKeyVaultClient
from azure.core.exceptions import ResourceNotFoundError

class _EnvVaraibleUtil(object):
    def __init__(self, config_path):
        self.config_path = config_path
        # Set to none first to avoid invalid reference
        self.akv_name = None
        self.akv_name = self.get_environment_variable_with_default( 'secrets', 'azure_key_vault', 'name')
        self.akv_client = AzureKeyVaultClient(self.akv_name) if self.akv_name else None

    def get_environment_variable_with_default(self, *args):
        """Gets the environment variable for the variable key.
        Args:
            *args: list of keys in feathr_config.yaml file
        Return:
            A environment variable for the variable key. It will retrieve the value of the environment variables in the following order:
            If the key is set in the environment variable, Feathr will use the value of that environment variable
            If it's not set in the environment, then a default is retrieved from the feathr_config.yaml file with the same config key.
            If it's not available in the feathr_config.yaml file, Feathr will try to retrieve the value from key vault
            If not found, an empty string will be returned with a warning error message.
            """

        # if envs exist, just return the existing env variable without reading the file
        env_keyword = "__".join(args)
        upper_env_keyword = env_keyword.upper()
        # make it work for lower case and upper case.
        env_variable = os.environ.get(
            env_keyword, os.environ.get(upper_env_keyword))

        # If the key is set in the environment variable, Feathr will use the value of that environment variable
        if env_variable:
            return env_variable

        # If it's not set in the environment, then a default is retrieved from the feathr_config.yaml file with the same config key.
        if os.path.exists(os.path.abspath(self.config_path)):
            with open(os.path.abspath(self.config_path), 'r') as stream:
                try:
                    yaml_config = yaml.safe_load(stream)
                    # concat all layers and  check in environment variable
                    yaml_layer = yaml_config

                    # resolve one layer after another
                    for arg in args:
                        yaml_layer = yaml_layer[arg]
                    return yaml_layer
                except KeyError as exc:
                    logger.info("{} not found in the config file.", env_keyword)
                except yaml.YAMLError as exc:
                    logger.warning(exc)
        
        # If it's not available in the feathr_config.yaml file, Feathr will try to retrieve the value from key vault
        if self.akv_name:
            try:
                return  self.akv_client.get_feathr_akv_secret(env_keyword) 
            except ResourceNotFoundError:
                # print out warning message if cannot find the env variable in all the resources
                logger.warning('Environment variable {} not found in environment variable, default YAML config file, or key vault service.', env_keyword)
                return None

    def get_environment_variable(self, variable_key):
        """Gets the environment variable for the variable key.
        

        Args:
            variable_key: environment variable key that is used to retrieve the environment variable
        Return:
            A environment variable for the variable key. It will retrieve the value of the environment variables in the following order:
            If the key is set in the environment variable, Feathr will use the value of that environment variable
            If it's not available in the environment variable file, Feathr will try to retrieve the value from key vault
            If not found, an empty string will be returned with a warning error message.
            """
        env_var_value = os.environ.get(variable_key)

        if env_var_value:
            return env_var_value

        # If it's not available in the environment variable file, Feathr will try to retrieve the value from key vault
        logger.info(variable_key + ' is not set in the environment variables.')
        
        if self.akv_name:
            try:
                return self.akv_client.get_feathr_akv_secret(variable_key)
            except ResourceNotFoundError:
                # print out warning message if cannot find the env variable in all the resources
                logger.warning('Environment variable {} not found in environment variable or key vault service.', variable_key)
                return None
    


from pprint import pprint
from typing import Union, List

from feathr.definition.anchor import FeatureAnchor
from feathr.definition.query_feature_list import FeatureQuery
from feathr.definition.materialization_settings import MaterializationSettings

class FeaturePrinter:
    """The class for pretty-printing features"""
    
    @staticmethod
    def pretty_print_anchors(anchor_list: List[FeatureAnchor]) -> None:
        """Pretty print features

        Args:
            feature_list: FeatureAnchor
        """

        if all(isinstance(anchor, FeatureAnchor) for anchor in anchor_list):
            for anchor in anchor_list:
                pprint("%s is the achor of %s" % \
                        (anchor.name, [feature.name for feature in anchor.features]))
        else:
            raise TypeError("anchor_list must be FeatureAnchor or List[FeatureAnchor]")

    @staticmethod
    def pretty_print_feature_query(feature_query: FeatureQuery) -> None:
        """Pretty print feature query

        Args:
            feature_query: feature query
        """

        if isinstance(feature_query, FeatureQuery):
            print("Features in feature_query: %s" % feature_query.feature_list)
        else:
            raise TypeError("feature_query must be FeatureQuery")

    @staticmethod
    def pretty_print_materialize_features(settings: MaterializationSettings) -> None:
        """Pretty print feature query

        Args:
            feature_query: feature query
        """
        if isinstance(settings, MaterializationSettings):
            print("Materialization features in settings: %s" % settings.feature_names)
        else:
            raise TypeError("settings must be MaterializationSettings")


import json
import os
import datetime
import time
from uuid import UUID
from feathr.definition.anchor import FeatureAnchor
from feathr.definition.dtype import BOOLEAN, FLOAT, INT32, BooleanFeatureType, ValueType
from feathr.definition.feature import Feature
from feathr.definition.feature_derivations import DerivedFeature
from feathr.definition.source import INPUT_CONTEXT, HdfsSource, InputContext
from feathr.definition.transformation import ExpressionTransformation, WindowAggTransformation
from feathr.definition.typed_key import TypedKey

from feathr.registry._feathr_registry_client import _FeatureRegistry, dict_to_source, dict_to_anchor, dict_to_feature, dict_to_derived_feature, dict_to_project


def test_parse_source():
    s = r'''{
        "attributes": {
            "name": "nycTaxiBatchSource",
            "path": "wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
            "preprocessing": "    def add_new_dropoff_and_fare_amount_column(df: DataFrame):\n        df = df.withColumn(\"new_lpep_dropoff_datetime\", col(\"lpep_dropoff_datetime\"))\n        df = df.withColumn(\"new_fare_amount\", col(\"fare_amount\") + 1000000)\n        return df\n",
            "qualifiedName": "feathr_getting_started__nycTaxiBatchSource",
            "tags": {
                "for_test_purpose": "true"
            },
            "type": "wasbs"
        },
        "displayText": "nycTaxiBatchSource",
        "guid": "c4a0ae0f-09cc-43bf-94e9-21ff178fbda6",
        "labels": [],
        "lastModifiedTS": "1",
        "name": "nycTaxiBatchSource",
        "qualifiedName": "feathr_getting_started__nycTaxiBatchSource",
        "status": "Active",
        "typeName": "feathr_source_v1"
    }'''
    source = dict_to_source(json.loads(s))
    assert isinstance(source, HdfsSource)
    assert source.name == "nycTaxiBatchSource"
    assert source.path == "wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv"
    assert source._registry_id == UUID("c4a0ae0f-09cc-43bf-94e9-21ff178fbda6")
    assert source._qualified_name == "feathr_getting_started__nycTaxiBatchSource"


def test_parse_anchor():
    s = r'''{
        "attributes": {
            "features": [
                {
                    "guid": "2380fe5b-ce2a-401e-98bf-af8b98460f67",
                    "typeName": "feathr_anchor_feature_v1",
                    "uniqueAttributes": {
                        "qualifiedName": "feathr_getting_started__request_features__f_day_of_week"
                    }
                },
                {
                    "guid": "103baca1-377a-4ddf-8429-5da91026c269",
                    "typeName": "feathr_anchor_feature_v1",
                    "uniqueAttributes": {
                        "qualifiedName": "feathr_getting_started__request_features__f_trip_time_duration"
                    }
                }
            ],
            "name": "request_features",
            "qualifiedName": "feathr_getting_started__request_features",
            "source": {
                "guid": "a4cfbc03-c65d-4f32-be3d-1d11247c9cdd",
                "typeName": "feathr_source_v1",
                "uniqueAttributes": {
                    "qualifiedName": "feathr_getting_started__PASSTHROUGH"
                }
            },
            "tags": {
                "for_test_purpose": "true"
            }
        },
        "displayText": "request_features",
        "guid": "260325a5-27f9-40d1-8697-c727feb1dbdc",
        "labels": [],
        "lastModifiedTS": "1",
        "name": "request_features",
        "qualifiedName": "feathr_getting_started__request_features",
        "status": "Active",
        "typeName": "feathr_anchor_v1"
    }'''
    anchor = dict_to_anchor(json.loads(s))
    # Parsed anchor is empty, features and source are added later
    assert anchor.name == "request_features"
    assert anchor._registry_id == UUID("260325a5-27f9-40d1-8697-c727feb1dbdc")
    assert anchor._qualified_name == "feathr_getting_started__request_features"


def test_parse_feature():
    s = r'''{
        "attributes": {
            "key": [
                {
                    "description": "A dummy typed key for passthrough/request feature.",
                    "fullName": "feathr.dummy_typedkey",
                    "keyColumn": "NOT_NEEDED",
                    "keyColumnAlias": "NOT_NEEDED",
                    "keyColumnType": "UNSPECIFIED"
                }
            ],
            "name": "f_is_long_trip_distance",
            "qualifiedName": "feathr_getting_started__request_features__f_is_long_trip_distance",
            "tags": {},
            "transformation": {
                "transformExpr": "cast_float(trip_distance)>30"
            },
            "type": {
                "dimensionType": [],
                "tensorCategory": "DENSE",
                "type": "TENSOR",
                "valType": "BOOLEAN"
            }
        },
        "displayText": "f_is_long_trip_distance",
        "guid": "dc24b1d5-206d-40db-b10a-606dd16a0297",
        "labels": [],
        "lastModifiedTS": "1",
        "name": "f_is_long_trip_distance",
        "qualifiedName": "feathr_getting_started__request_features__f_is_long_trip_distance",
        "status": "Active",
        "typeName": "feathr_anchor_feature_v1"
    }'''
    f: Feature = dict_to_feature(json.loads(s))
    assert f.feature_type == BOOLEAN
    assert f.name == "f_is_long_trip_distance"
    assert len(f.key) == 1
    assert f.key[0].key_column_type == ValueType.UNSPECIFIED
    assert f.key[0].key_column == "NOT_NEEDED"
    assert isinstance(f.transform, ExpressionTransformation)
    assert f.transform.expr == "cast_float(trip_distance)>30"
    # Hidden properties
    assert f._qualified_name == "feathr_getting_started__request_features__f_is_long_trip_distance"
    assert f._registry_id == UUID("dc24b1d5-206d-40db-b10a-606dd16a0297")


def test_parse_derived_feature():
    s = r'''{
        "attributes": {
            "inputAnchorFeatures": [],
            "inputDerivedFeatures": [
                {
                    "guid": "c626c41c-d6c2-4b16-a267-6cdeea497c52",
                    "typeName": "feathr_anchor_feature_v1",
                    "uniqueAttributes": {
                        "qualifiedName": "feathr_getting_started__f_trip_time_rounded"
                    }
                }
            ],
            "key": [
                {
                    "description": "A dummy typed key for passthrough/request feature.",
                    "fullName": "feathr.dummy_typedkey",
                    "keyColumn": "NOT_NEEDED",
                    "keyColumnAlias": "NOT_NEEDED",
                    "keyColumnType": "UNSPECIFIED"
                }
            ],
            "name": "f_trip_time_rounded_plus",
            "qualifiedName": "feathr_getting_started__f_trip_time_rounded_plus",
            "tags": {},
            "transformation": {
                "transformExpr": "f_trip_time_rounded + 100"
            },
            "type": {
                "dimensionType": [],
                "tensorCategory": "DENSE",
                "type": "TENSOR",
                "valType": "INT"
            }
        },
        "displayText": "f_trip_time_rounded_plus",
        "guid": "479c6306-5fdb-4e06-9008-c18f68db52a4",
        "labels": [],
        "lastModifiedTS": "1",
        "name": "f_trip_time_rounded_plus",
        "qualifiedName": "feathr_getting_started__f_trip_time_rounded_plus",
        "status": "Active",
        "typeName": "feathr_derived_feature_v1"
    }'''
    df = dict_to_derived_feature(json.loads(s))
    assert df.name == "f_trip_time_rounded_plus"
    assert df.feature_type == INT32
    assert len(df.key) == 1
    assert df.key[0].key_column_type == ValueType.UNSPECIFIED
    # Hidden properties
    assert df._qualified_name == "feathr_getting_started__f_trip_time_rounded_plus"
    assert df._registry_id == UUID("479c6306-5fdb-4e06-9008-c18f68db52a4")

def test_parse_project():
    filename = os.path.join(os.path.dirname(__file__), "test_registry_lineage.json")
    f=open(filename, "r")
    (anchors, derived_features) = dict_to_project(json.load(f))
    assert len(anchors)==2
    request_features = [a for a in anchors if a.name=='request_features'][0]
    assert isinstance(request_features.source, InputContext)
    assert len(request_features.features)==4
    aggregationFeatures = [a for a in anchors if a.name=='aggregationFeatures'][0]
    assert len(aggregationFeatures.features)==2
    assert isinstance(aggregationFeatures.source, HdfsSource)
    assert aggregationFeatures.source.name=="nycTaxiBatchSource"
    assert aggregationFeatures.source.path=="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv"
    assert len(derived_features)==3

def test_registry_client_list_features():
    c = _FeatureRegistry(project_name="p", endpoint="https://feathr-sql-registry.azurewebsites.net/api/v1")
    f = [e["qualifiedName"] for e in c.list_registered_features("feathr_ci_registry_getting_started")]
    print(f)
    assert len(f)==9
    for i in f:
        assert i.startswith("feathr_ci_registry_getting_started__")
        
def test_registry_client_load():
    c = _FeatureRegistry(project_name="p", endpoint="https://feathr-sql-registry.azurewebsites.net/api/v1")
    (anchors, derived_features) = c.get_features_from_registry("feathr_ci_registry_getting_started")
    assert len(anchors)==2
    request_features = [a for a in anchors if a.name=='request_features'][0]
    assert isinstance(request_features.source, InputContext)
    assert len(request_features.features)==4
    aggregationFeatures = [a for a in anchors if a.name=='aggregationFeatures'][0]
    assert len(aggregationFeatures.features)==3
    assert isinstance(aggregationFeatures.source, HdfsSource)
    assert aggregationFeatures.source.name=="nycTaxiBatchSource"
    assert aggregationFeatures.source.path=="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04_with_index.csv"
    assert len(derived_features)==2

def test_create():
    project_name = f"feathr_registry_client_test_{int(time.time())}"
    c = _FeatureRegistry(project_name="p", endpoint="https://feathr-sql-registry.azurewebsites.net/api/v1")
    
    batch_source = HdfsSource(name="nycTaxiBatchSource",
                            path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                            event_timestamp_column="lpep_dropoff_datetime",
                            timestamp_format="yyyy-MM-dd HH:mm:ss")

    f_trip_distance = Feature(name="f_trip_distance",
                              feature_type=FLOAT, transform="trip_distance")
    f_trip_time_duration = Feature(name="f_trip_time_duration",
                                   feature_type=INT32,
                                   transform="(to_unix_timestamp(lpep_dropoff_datetime) - to_unix_timestamp(lpep_pickup_datetime))/60")

    features = [
        f_trip_distance,
        f_trip_time_duration,
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]


    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features)

    f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                                          feature_type=FLOAT,
                                          input_features=[
                                              f_trip_distance, f_trip_time_duration],
                                          transform="f_trip_distance * f_trip_time_duration")

    f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_duration],
                                         transform="f_trip_time_duration % 10")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")
    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="AVG",
                                                              window="90d")),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)
    
    c.register_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=[f_trip_time_distance, f_trip_time_rounded])

    

if __name__ == "__main__":
    test_parse_source()
    test_parse_anchor()
    test_parse_feature()
    test_parse_derived_feature()
    test_parse_project()
    test_registry_client_list_features()
    test_registry_client_load()
    test_create()



import sys
import os
from setuptools import setup, find_packages
from pathlib import Path

# Use the README.md from /docs
root_path = Path(__file__).resolve().parent.parent
long_description = (root_path / "docs/README.md").read_text(encoding="utf8")

try:
    exec(open("feathr/version.py").read())
except IOError:
    print("Failed to load Feathr version file for packaging.",
          file=sys.stderr)
    sys.exit(-1)

VERSION = __version__  # noqa
os.environ["FEATHR_VERSION]"] = VERSION

extras_require=dict(
    dev=[
        "black>=22.1.0",    # formatter
        "isort",            # sort import statements
        "pytest>=7",
        "pytest-xdist",
        "pytest-mock>=3.8.1",
    ],
    notebook=[
        "jupyter==1.0.0",
        "matplotlib==3.6.1",
        "papermill>=2.1.2,<3",      # to test run notebooks
        "scrapbook>=0.5.0,<1.0.0",  # to scrap notebook outputs
    ],
)
extras_require["all"] = list(set(sum([*extras_require.values()], [])))

setup(
    name='feathr',
    version=VERSION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email="feathr-technical-discuss@lists.lfaidata.foundation",
    description="An Enterprise-Grade, High Performance Feature Store",
    url="https://github.com/feathr-ai/feathr",
    project_urls={
        "Bug Tracker": "https://github.com/feathr-ai/feathr/issues",
    },
    packages=find_packages(),
    include_package_data=True,
    # consider
    install_requires=[
        "click<=8.1.3",
        "py4j<=0.10.9.7",
        "loguru<=0.6.0",
        "pandas<=1.5.0",
        "redis<=4.4.0",
        "requests<=2.28.1",
        "tqdm<=4.64.1",
        "pyapacheatlas<=0.14.0",
        "pyhocon<=0.3.59",
        "pandavro<=1.7.1",
        "pyyaml<=6.0",
        "Jinja2<=3.1.2",
        "pyarrow<=9.0.0",
        "pyspark>=3.1.2",
        "python-snappy<=0.6.1",
        "deltalake>=0.6.2",
        "graphlib_backport<=1.0.3",
        "protobuf<=3.19.4,>=3.0.0",
        "confluent-kafka<=1.9.2",
        "databricks-cli<=0.17.3",
        "avro<=1.11.1",
        "azure-storage-file-datalake<=12.5.0",
        "azure-synapse-spark<=0.7.0",
        # fixing Azure Machine Learning authentication issue per https://stackoverflow.com/a/72262694/3193073
        "azure-identity>=1.8.0",
        "azure-keyvault-secrets<=4.6.0",
        # In 1.23.0, azure-core is using ParamSpec which might cause issues in some of the databricks runtime.
        # see this for more details:
        # https://github.com/Azure/azure-sdk-for-python/pull/22891
        # using a version lower than that to workaround this issue.
        "azure-core<=1.22.1",
        # azure-core 1.22.1 is dependent on msrest==0.6.21, if an environment(AML) has a different version of azure-core (say 1.24.0),
        # it brings a different version of msrest(0.7.0) which is incompatible with azure-core==1.22.1. Hence we need to pin it.
        # See this for more details: https://github.com/Azure/azure-sdk-for-python/issues/24765
        "msrest<=0.6.21",
        "typing_extensions>=4.2.0"
    ],
    tests_require=[  # TODO: This has been depricated
        "pytest",
    ],
    extras_require=extras_require,
    entry_points={
        'console_scripts': ['feathr=feathrcli.cli:cli']
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7"
)



from feathrcli.cli import init
from click.testing import CliRunner
import os
from feathr import FeathrClient


def test_configuration_loading():
    """
    Test the configuration can be overwritten by envs
    """
    runner = CliRunner()
    with runner.isolated_filesystem():

        result = runner.invoke(init, [])

        assert result.exit_code == 0
        assert os.path.isdir('./feathr_user_workspace')

        client = FeathrClient(config_path='./feathr_user_workspace/feathr_config.yaml')

        # test the loading is correct even if we are not in that folder
        assert client._FEATHR_JOB_JAR_PATH is not None

        SPARK_RESULT_OUTPUT_PARTS = '4'

        # Use a less impactful config to test, as this config might be impactful for all the tests (since it's setting the envs)
        os.environ['SPARK_CONFIG__SPARK_RESULT_OUTPUT_PARTS'] = SPARK_RESULT_OUTPUT_PARTS

        # this should not be error out as we will just give users prompt, though the config is not really here
        client = FeathrClient(config_path='./feathr_user_workspace/feathr_config.yaml')
        assert client.output_num_parts == SPARK_RESULT_OUTPUT_PARTS
